//      check_metafs.c
//      
//      Copyright 2010 Leszek Kattinger <letesch@gmx.de>
//      
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//      
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//      
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.


#define _GNU_SOURCE
#define _XOPEN_SOURCE 500
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>
#include <grp.h>
#include <time.h>
#include <locale.h>
#include <langinfo.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <glib.h>
//#include <pthread.h>
#include <assert.h>
#include <signal.h>
#include "metafs.h" 

/*
gcc `pkg-config --cflags --libs glib-2.0` -lpthread metafs.c check_metafs.c -o check_metafs

run parameter: <filename> <testnr>
*/

// nlen + mode + uid + gid + times + data
// 1 		+ 4  	 + 4 	 + 4	 +	24	 +	8	= 45

#define CONST_PHYSICAL_ENTRY_SIZE 45

// fillpos + ext_block_nr = 4	+ 8 = 12

#define CONST_BLOCK_HEADER_SIZE 12 

static void signalHandler(int signal)
{
  printf("Caught signal %d\n", signal);  
 
	int c;

  if(signal != SIGINT)
      return;
  else {
      printf("\n Quit ? (j/N) : ");
      c=getchar();
      if(c == 'j')				 
         exit (EXIT_FAILURE);
      else
         return;
   }     
}


int main ( int argc, char *argv[] )
{  			  		
	if( argc < 3)
	{
		printf ("Enter Filesystem location and select test case! \n \
            check_metafs <file/device> < 1 | 2 | 3 | 4>  \n");
		return false;
	}	
	
	signal(SIGINT, & signalHandler);
		
	Super * s = (Super *)initSuper();	
		
	s->file  = open(argv[1], O_RDONLY  );
	
	if( s->file  == -1 ) {
		fprintf(stderr, "Problem format device - can't open  %s \n", argv[1] );
    return false;	}	
		
	readSuper(s);	
	initBitmap(s);
	calculateSuper(s);			
				
	puts("-----------------------------------");
	printf("Test Filesystem MetaFS on %s  \n", argv[1] );
	puts("-----------------------------------");
	
	readBitmap(s);
	
	displaySuper(s);		
	
	int select = 1;   // default
  
	select = atoi(argv[2]);			
		
	puts("-----------------------------------------------------------------------------------------------------");
	switch(select) 
		{
			case 1 : 		 
				printf("Start read allocated block-header: block load in [%%] on:  %s \n", argv[1] );
			break;  
				
			case 2 : 						
				printf("Start read allocated full-blocks: block read time in [ms] on:  %s \n", argv[1] );
      break;  
					
			case 3 : 						
			  printf("Start read allocated block-header + content: block read time in [ms] on:  %s \n", argv[1] );
			break;  	
					
			case 4 : 						
			  printf("Start read random blocks: block read time in [ms] on:  %s \n", argv[1] );
			break; 								
		}
	puts("-----------------------------------------------------------------------------------------------------");					
	uint64_t metadata_size = (uint64_t)(50);		// avoid division by 0 for empty fs (root entry metadata)
	
	double percent =(double) (0);		
		
	uint32_t fillpos = 0;
	
	uint64_t block_count = 0 , tmp_block = 0;
	
	struct timeval start, stop;
	long long unsigned count_time_usec = 0;	
	long long unsigned count_time_sec = 0;		
	long long unsigned tmp_lapsed_usec = 0;
	long long unsigned tmp_lapsed_sec = 0;  
	
	long int n;	
	  	
	char buff[s->block_size + 1]; 

	ssize_t ret;
	
  uint64_t i, j;
  	
	for( i = 0; i < s->bitmap_size; i++)		// iterate bitmap to find used blocks -> only they have a valid fillpos!
	{
		tmp_block = (uint64_t)(i * 32);		// set tmp block_nr to byte_pos 
		
		fillpos = 0;  // reset fillpos
	
		j = (i) ? 0 : 2;  					// skip super block and bitmap
				
		for( ; j<32; j++)						// iterate bitmap unit 4 Byte
		{						
			if( ((s->bitmap[i] >> j) % 2) > 0) 	// find used Bit = 1 in bitmap unit 
			{ 
				block_count++;										// count used blocks	
				
				n = random();	
				
				n = n % s->blocks;		
								
		    gettimeofday (&start, NULL);	// start time 
				
				switch(select) 
				{
					case 1 : 				
						
						ret = pread(s->file, &fillpos, 4 , (tmp_block+j) * s->block_size);		// read block-header			
						if(ret < 4)	{
							fprintf(stderr, "Problem read bs - read Bytes %llu @ block %"PRIu64"\n", (long long unsigned) ret, tmp_block+j );
							return false; }	
						metadata_size += fillpos;			// filesystem metadata count in Bytes
					break;
		  	 
					case 2 : 				 
		  	 	
						ret = pread(s->file, buff, (size_t)(s->block_size) , (tmp_block+j) * s->block_size);		// read block-content	
						if(ret < s->block_size)	{
							fprintf(stderr, "Problem read bs - read Bytes %llu @ block %"PRIu64"\n", (long long unsigned) ret, tmp_block+j );
							return false; }							 		
					break;	
					
					case 3 : 				 
		  	 	
						ret = pread(s->file, &fillpos, 4 , (tmp_block+j) * s->block_size);		// read block-header			
						if(ret < 4)	{
							fprintf(stderr, "Problem read bs - read Bytes %llu @ block %"PRIu64"\n", (long long unsigned) ret, tmp_block+j );
							return false; }	
						metadata_size += fillpos;			// filesystem metadata count in Bytes
						
						ret = pread(s->file, buff, (size_t)(fillpos) , (tmp_block+j) * s->block_size+12);		// read block-content	
						if(ret < fillpos)	{
							fprintf(stderr, "Problem read bs - read Bytes %llu @ block %"PRIu64"\n", (long long unsigned) ret, tmp_block+j );
							return false; }							 		
					break;		
					
					case 4 :
											
						ret = pread(s->file, buff, (size_t)(s->block_size) , n * s->block_size);		// read full-block
						if(ret < s->block_size)	{
							fprintf(stderr, "Problem read bs - read Bytes %llu @ block %li\n", (long long unsigned) ret, n );
							return false; }							 		
					break;													
				}							
				
				gettimeofday (&stop, NULL);				// stop time 
				
				if (start.tv_usec > stop.tv_usec) 
				{  
					stop.tv_usec += 1000000;  stop.tv_sec--; 
				} 					
								
				tmp_lapsed_usec += (stop.tv_usec - start.tv_usec);
				tmp_lapsed_sec += (stop.tv_sec - start.tv_sec);
				
				count_time_usec += tmp_lapsed_usec;
				count_time_sec += tmp_lapsed_sec; 				
			 		
			 switch(select) 
       {
         case 1 : 		
							
           percent = (double)( (fillpos * (uint32_t)(100)) / s->block_size );	
		  									
					 if( percent > (double)1.0 )									
             printf(" %2.0lf", percent ); 
           else printf("  1"); 
						 
           if( (block_count % 32) == 0 )	printf("\n");						
				 break;  
					
				 case 2 :
				
				   printf(" %llu", tmp_lapsed_usec  / 1000  );
				 
				   if( (block_count % 32) == 0 )	printf("\n");						 
         break;  				
				
         case 3 :
				
					 percent = (double)( (fillpos * 100) / s->block_size );	
					
					 printf("|" );	
					 
					 if( (block_count % 32) == 0 )	printf("\r");	
         break;
					
         case 4 :
				
				   printf(" %llu", tmp_lapsed_usec / 1000  );
				 
				   if( (block_count % 32) == 0 )	printf("\n");						 
				 break;  						
				
       }							 			
       tmp_lapsed_usec = tmp_lapsed_sec = 0;  // reset time				
										
			}			
		}			  
	}		
  
	printf (" \n done! \n");	
	
	uint64_t volume_size_mb = s->volume_size/(1024*1024);
	
	double metadata_size_mb = 	(double)(metadata_size )/ (1024*1024) ;		
	
	long long unsigned count_time_msec = count_time_usec / 1000 ; 
	
	count_time_sec = 	count_time_msec / 1000; 
	
  switch(select) 
	{
	 	case 1 : 	
				
		  percent =  (double)(metadata_size * 100 )  / s->volume_size  ;	
							
			puts("-----------------------------------------------------------------------------------------------------------");
			printf (" Stored metadata size: %"PRIu64" Byte [%5.2lf MB] - Space used: %3.2lf %%  of %"PRIu64" MB \n",
  																				metadata_size, metadata_size_mb, 	percent,  volume_size_mb );	
			puts("-----------------------------------------------------------------------------------------------------------");	
			printf (" \n Read %"PRIu64" block-headers in %llus %llums  ", block_count, 	count_time_sec, count_time_msec % 1000); 			
				
			printf(" => average block-header read-time: %llu us/block\n" , count_time_usec/block_count ); 		
    break;
          
    case 2 :    
            
		  printf (" \n Read %"PRIu64" blocks in in %llus %llums  ", block_count, 	count_time_sec, count_time_msec % 1000); 			
					
			printf(" => average block read-time: %llu us/block \n" , count_time_usec/block_count);  					 
		break;
					
		case 3 :   
					
			percent =  (double)(metadata_size * 100 )  / s->volume_size  ;	
					
			puts("-----------------------------------------------------------------------------------------------------------");
			printf (" Stored metadata size: %"PRIu64" Byte [%5.2lf MB] - Space used: %3.2lf %%  of %"PRIu64" MB \n",
  																				metadata_size, metadata_size_mb, 	percent,  volume_size_mb );	
			puts("-----------------------------------------------------------------------------------------------------------");						
      printf (" \n Read %"PRIu64" block-headers and block-fillpos in %llus %llums  ", block_count, count_time_sec, count_time_msec % 1000); 			
					
			printf(" => average  read-time: %llu us/block \n" , count_time_usec/block_count);  											  
		break;
						
    case 4 :   
					
			printf (" \n Read %"PRIu64" random blocks in in %llus %llums  ", block_count, 	count_time_sec, count_time_msec % 1000); 			
					
			printf(" => average random block [%"PRIu32"] read-time: %llu us/block \n" ,s->block_size, count_time_usec/block_count);  					 						  
		break;							
  }						
		
	close(s->file);		
	
  return 0;
}









