#include <stdbool.h>     
#include <glib.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <utime.h>
#include <inttypes.h>
#include <pthread.h>

// nlen + mode + uid + gid + times + data
//  1		+  4 	 +  4  +  4	 +	24	 +	8	= 48

#define CONST_PHYSICAL_ENTRY_SIZE 45

// fillpos + ext_block_nr
//		 4   +     8        = 12

#define CONST_BLOCK_HEADER_SIZE 12 

typedef struct entry
{
  char * name;				// dir or filename  
  
  uint8_t name_len;		// filename length without /0  
	
  uint64_t data;			// do it later 
	/*
	union data{	// content of file -> data or directory -> entries  
    uint64_t file_block;  	
    uint64_t first_directory_block;
  }
	*/
  struct stat a;						// System POSIX stat struct 
	
	struct block * block_ptr; // shortcut to block node
	
	struct entry * next;
	struct entry * prev;	

}Entry;
		
	
typedef struct blockstat
{
  uint32_t  fillpos;				// fill level of bock
	
  uint64_t  ext_block_nr;		// extent block number

}Blockstat;

	
typedef struct block
{
  uint64_t  block_nr;
        
  uint8_t	  dirty; 	// 0=clean 
	
	Blockstat * bs;		// fillpos and ext_block_number
     
	Entry * e;
	Entry * last_e;
    			
  struct dir * dir_ptr;	// back pointer to dir structure 
    
	struct block * next;
	struct block * prev;
		
}Block;


typedef struct dir
{
	Block * b;				// pointer to first block list node 
	Block * last_b;		// pointer to last block list node
	
	uint8_t	 dirty; 	// not used now 	
	
	uint32_t fillpos;	// not used now
	
	uint64_t dir_nr; 	// first block_nr -> dir_nr / physical <-> logical
	
	GTree * t;				// tree with all dentries of this directory
  					
}Dir;


typedef struct position
{
	Dir * d;	//parent Dir 
	
	Entry * e;

}Position;


typedef struct super
{
	uint8_t	 dirty; 				// 0=clean 

	uint32_t bitmap_size;		// in units (a 4Byte)
	
	uint32_t bitmap_bsize;	// in bytes
	
	uint32_t * bitmap;			// pointer to bitmap in memory
	
  uint32_t block_size;		// block size in bytes
	
	uint64_t volume_size;		// size of the physical filesystem file / device
		
	uint64_t fillpos;				// sum of all loaded fill positions 	
	
	uint64_t blocks;				// volume block count witout offcut	
	
	uint64_t free_blocks;
	
	uint64_t used_blocks;
		
	Entry * root;						// pointer to the root entry	
	
	int file;								// metadata volume file descriptor int
	
	int data_file;					// log file descriptor int
	
	uint64_t bitmap_backup_block;
	
	uint64_t super_backup_block;
	
	//char fs_name[] = "metafs";		// fs check by name
							
}Super;


typedef struct {  	
	
  Super * s;		// super block 
	
	Dir * d;			// pointer to the loaded root dir structure / content

	GTree * dt;		// tree with laded directories	
	
  bool closed;	// filesystem status
	
	GList * dirty_block_list;				// list with modified blocks
   	
	pthread_mutex_t  write_mutex;		// write back mutex
	pthread_cond_t   enforce_write_back_cond;
	
  pthread_t write_back_thread; 
  
  uint64_t loaded_byte_count;
  uint64_t max_cached_byte_count;
        
	// anzahl dirty Blöcke
	unsigned dirty_block_count;
	// anzahl der geladenen Blöcke
	unsigned loaded_block_count;	
	// maximum number of cachable blocks
	unsigned max_cached_block_count;
			
	// erstmal gilt 0 < dirty_block_count <= loaded_block_count <= max_cached_block_count

	// Wann stoße ich den background_writer thread zusätzlich an?
	// Wenn kein/wenig Platz für neue Dirs ist und eine großzahl der Dirs dirty ist!
	// => if (((float) loaded_block_count) / max_cached_block_count > 0.8 && ((float) dirty_block_count) / loaded_block_count) > 0.5) => wakeup background thread 
	// Dies muss geprüft werden (mit Mutex) wann ich einen Block dirty mache!!!
} Filesystem;

extern int dbg;	// debug level 0 - 3

// Filesystem Usage API

extern Filesystem mefs_loadFilesystem(char * metadataFile, char * dataFile);

extern bool mefs_destroy(); 

extern bool mefs_create(Super * s, Dir * d, Entry * e);

extern bool mefs_rename(Super * s, Entry * e);

extern bool mefs_delete(Super * s, Entry * e);

extern bool mefs_chmod(Entry * e, mode_t mode);

extern bool mefs_chown(Entry * e, uid_t uid, gid_t gid);

extern bool mefs_utimens(Entry * e, struct timeval ts[2]);




// Filesystem Creation API

extern Super * initSuper();

extern bool calculateSuper(Super * s);

extern bool createSuper(Super * s);

extern bool readSuper(Super * s);

extern bool writeSuper(Super * s);

extern bool displaySuper(Super * s);


extern bool initBitmap(Super * s);

extern bool readBitmap(Super * s);

extern bool writeBitmap(Super * s);

extern bool displayBitmap(Super * e);


extern bool displayEntry(Entry * e);

extern bool displayBlockStat(Block * b);





///////////////////////////////////////////

extern bool setBlockUsed_noLock(Super * s, uint64_t block_nr);

extern bool setBlockFree_noLock(Super * s, uint64_t block_nr);

extern uint64_t getFreeBlock_noLock(Super * s);

extern bool appendDir( Dir * d);

extern bool appendBlock(Dir * d, Block * b);

extern bool appendEntry(Block * b, Entry * e);
 
extern bool compressBlocks(Super * s, Dir * d);

extern bool removeDir(Super * s, uint64_t dir_nr);

extern bool removeEntry(Dir * d, Entry * e);

extern Block * initBlock(uint64_t block_nr);

extern Block * loadBlock(uint64_t block_nr);

extern Blockstat * readBlockStat(Super * s, uint64_t block_nr);

extern bool readBlockEntries(Block * b, char * buff);

extern uint32_t readFillpos(Super * s, uint64_t block_nr);

extern bool appendBlock(Dir * d, Block * b);

extern bool unloadBlock(Block * b);

extern bool removeBlock(Block * b);



extern bool statBlockToBuff(Block * b, char * buff, off_t * curBuffPos);

extern bool entryToBuff(Entry * e, char * buff, off_t * curBuffPos);

extern bool buffToEntry(Entry * e, char * buff, off_t * curBuffPos);

extern bool superToBuff(Super * s, char * buff, off_t * curBuffPos);

extern bool buffToSuper(Super * s, char * buff, off_t * curBuffPos);



extern bool loadTree(Dir * d);

extern Dir * initDir(uint64_t block_nr);

extern Dir * createDir_noLock(uint64_t block_nr);

extern bool unloadDir(Dir * d);

extern bool appendDir( Dir * d);



extern gint block_nr_cmp(const guint64 * a, const guint64 * b);



extern bool initLoadedDirsTree();

extern bool unloadLoadedDirs();

extern GList * buildLookupList(const char * path);

extern gboolean iter_all_loaded_dirs( gpointer nr, Dir * d);


extern Position * new_mefs_lookupPath(const char * path); 

extern Dir * mefs_loadDir(uint64_t block_nr);

extern Dir * mefs_lookupDir(uint64_t dir_nr);

extern Entry * mefs_lookupEntry(Dir * d, char * name);

extern Entry * mefs_lookupPath(const char * path);

extern Entry * mefs_initEntry(char * name, mode_t mode);
 
 
 





















